from collections import OrderedDict
from typing import Optional, Sequence, Union

from brownie.utils.color import Color

color = Color()


def build_tree(
    tree_dict: Union[OrderedDict, Sequence],
    multiline_pad: int = 1,
    _indent_data: Optional[list] = None,
) -> str:
    """
    Build a tree graph from a nested OrderedDict.

    Each dictionary key is a value to be added to the tree. The value may be:

    * an OrderedDict, containing keys to be placed beneath it within the tree
    * a List, for keys beneath it in the node when none also have their own sub-nodes
    * None, when the key has no subnode

    For keys that contain a new line, all lines beyond the first are indented. It
    is possible to create complex trees that contain subtrees, by using the tree_str
    of `build_tree` as a key value in another tree.

    Arguments
    ---------
    tree_dict : OrderedDict
        OrderedDict to be turned into a tree.
    multiline_pad : int, optional
        Number of padding lines to leave befor and after a tree value that spans
        multiple lines.
    _indent_data
        Internal list to handle indentation during recursive calls. The initial
        call to this function should always leave this value as `None`.

    Returns
    -------
    str
        Tree graph.
    """
    tree_str = f"{color('dark white')}"
    if _indent_data is None:
        _indent_data = []

    was_padded = False
    for i, key in enumerate(tree_dict):
        is_last_item = bool(i < len(tree_dict) - 1)

        # create indentation string
        indent = f"{color('dark white')}"
        for value in _indent_data[1:]:
            indent = f"{indent}\u2502   " if value else f"{indent}    "
        if _indent_data:
            symbol = "\u251c" if is_last_item else "\u2514"
            indent = f"{indent}{symbol}\u2500\u2500 "

        lines = [i for i in key.split("\n") if i]
        if len(lines) > 1 and not was_padded:
            for i in range(multiline_pad):
                tree_str = f"{tree_str}{indent[:-4]}\u2502   \n"

        tree_str = f"{tree_str}{indent}{color}{lines[0]}\n"
        was_padded = False

        if len(lines) > 1:
            # handle multiline keys
            symbol = "\u2502" if is_last_item else " "
            symbol2 = "\u2502" if (isinstance(tree_dict, dict) and tree_dict[key]) else " "
            indent = f"{indent[:-4]}{symbol}   {symbol2}   "
            for line in lines[1:] + ([""] * multiline_pad):
                tree_str = f"{tree_str}{color('dark white')}{indent}{color}{line}\n"
            was_padded = True

        if isinstance(tree_dict, dict) and tree_dict[key]:
            # create nested tree
            nested_tree = build_tree(tree_dict[key], multiline_pad, _indent_data + [is_last_item])
            tree_str = f"{tree_str}{nested_tree}"

    return tree_str



#!/usr/bin/python3

import json
from hashlib import sha1
from pathlib import Path

import hypothesis
from py.path import local

import brownie
from brownie._config import CONFIG
from brownie.project.scripts import _get_ast_hash
from brownie.test import _apply_given_wrapper, coverage, output

from .utils import convert_outcome


class PytestBrownieBase:
    """
    Brownie plugin base hooks.

    Pytest hooks in this class are used in every testing mode.
    """

    def __init__(self, config, project):
        _apply_given_wrapper()

        self.config = config
        # required when brownie project is in a subfolder of another project
        config.rootdir = local(project._path)

        self.project = project
        self.project_path = project._path

        self.results = {}
        self.node_map = {}
        self.isolated = {}
        self.skip = {}
        self.contracts = dict(
            (k, v["bytecodeSha1"]) for k, v in project._build.items() if v.get("bytecode")
        )

        glob = self.project_path.joinpath(self.project._structure["tests"]).glob("**/conftest.py")
        self.conf_hashes = dict((self._path(i.parent), _get_ast_hash(i)) for i in glob)
        try:
            with self.project._build_path.joinpath("tests.json").open() as fp:
                hashes = json.load(fp)
        except (FileNotFoundError, json.decoder.JSONDecodeError):
            hashes = {"tests": {}, "contracts": {}, "tx": {}}

        self.tests = dict(
            (k, v)
            for k, v in hashes["tests"].items()
            if self.project_path.joinpath(k).exists() and self._get_hash(k) == v["sha1"]
        )

        changed_contracts = set(
            k
            for k, v in hashes["contracts"].items()
            if k not in self.contracts or v != self.contracts[k]
        )
        if changed_contracts:
            for txhash, coverage_eval in hashes["tx"].items():
                if not changed_contracts.intersection(coverage_eval.keys()):
                    coverage._add_cached_transaction(txhash, coverage_eval)
            self.tests = dict(
                (k, v)
                for k, v in self.tests.items()
                if v["isolated"] is not False and not changed_contracts.intersection(v["isolated"])
            )
        else:
            for txhash, coverage_eval in hashes["tx"].items():
                coverage._add_cached_transaction(txhash, coverage_eval)

    def _reduce_path_strings(self, text):
        # convert absolute path strings to relative ones, prior to outputting to console
        base_path = f"{Path(brownie.__file__).parent.as_posix()}"
        project_path = f"{self.project_path.as_posix()}/"

        text = text.replace(base_path, "brownie")
        text = text.replace(project_path, "")
        return text

    def _path(self, path):
        return self.project_path.joinpath(path).relative_to(self.project_path).as_posix()

    def _test_id(self, nodeid):
        path, test_id = nodeid.split("::", maxsplit=1)
        return self._path(path), test_id

    def _get_hash(self, path):
        hash_ = _get_ast_hash(self.project_path.joinpath(path))
        for confpath in filter(lambda k: k in path, sorted(self.conf_hashes)):
            hash_ += self.conf_hashes[confpath]
        return sha1(hash_.encode()).hexdigest()

    def pytest_configure(self, config):
        for key in ("coverage", "always_transact"):
            CONFIG.argv[key] = config.getoption("--coverage")
        CONFIG.argv["cli"] = "test"
        CONFIG.argv["gas"] = config.getoption("--gas")
        CONFIG.argv["revert"] = config.getoption("--revert-tb")
        CONFIG.argv["update"] = config.getoption("--update")
        CONFIG.argv["network"] = None
        if config.getoption("--network"):
            CONFIG.argv["network"] = config.getoption("--network")[0]

    def _make_nodemap(self, ids):
        self.node_map.clear()
        for item in ids:
            path, test = self._test_id(item)
            self.node_map.setdefault(path, []).append(test)

    def pytest_sessionstart(self):
        """
        Called after the `Session` object has been created and before performing
        collection and entering the run test loop.

        * Replaces the default hypothesis reporter with a one that applies source
          highlights and increased vertical space between results. The effect is
          seen in output for `hypothesis.errors.MultipleFailures` and while the
          `-v` flag is active.

        * Removes `PytestAssertRewriteWarning` warnings from the terminalreporter.
          This prevents warnings that "the `brownie` library was already imported and
          so related assertions cannot be rewritten". The warning is not relevant
          for end users who are performing tests with brownie, not on brownie,
          so we suppress it to avoid confusion.

        Removal of pytest warnings must be handled in this hook because session
        information is passed between xdist workers and master prior to test execution.
        """

        def _hypothesis_reporter(text):
            text = self._reduce_path_strings(text)
            if next((i for i in ("Falsifying", "Trying", "Traceback") if text.startswith(i)), None):
                print("")

            lines = [
                reporter._tw._highlight(i) if not i.lstrip().startswith("\x1b") else f"{i}\n"
                for i in text.split("\n")
            ]
            text = "".join(lines)

            end = "\n" if text.startswith("Traceback") else ""
            print(text, end=end)

        hypothesis.reporting.reporter.default = _hypothesis_reporter
        hypothesis.extra.pytestplugin.default_reporter = _hypothesis_reporter

        reporter = self.config.pluginmanager.get_plugin("terminalreporter")
        warnings = reporter.stats.pop("warnings", [])
        warnings = [i for i in warnings if "PytestAssertRewriteWarning" not in i.message]
        if warnings and not self.config.getoption("--disable-warnings"):
            reporter.stats["warnings"] = warnings

    def pytest_report_teststatus(self, report):
        """
        Return result-category, shortletter and verbose word for status reporting.

        With the `--update` flag, modifies the outcome of already-run skipped
        tests so that the final report shows accurate pass/fail information.

        Arguments
        ---------
        report : _pytest.reports.BaseReport
            Report object for the current test.
        """
        if report.when == "setup":
            self.skip[report.nodeid] = report.skipped
            if report.failed:
                return "error", "E", "ERROR"
            return "", "", ""
        if report.when == "teardown":
            if report.failed:
                return "error", "E", "ERROR"
            elif self.skip[report.nodeid]:
                path, test_id = self._test_id(report.nodeid)
                idx = self.node_map[path].index(test_id)
                report.outcome = convert_outcome(self.results[path][idx])
                return "skipped", "s", "SKIPPED"
            return "", "", ""
        if hasattr(report, "wasxfail"):
            if report.skipped:
                return "xfailed", "x", "XFAIL"
            elif report.passed:
                return "xpassed", "X", "XPASS"
        return report.outcome, convert_outcome(report.outcome), report.outcome.upper()

    def pytest_runtest_makereport(self, item):
        """
        Return a _pytest.runner.TestReport object for the given pytest.Item and
        _pytest.runner.CallInfo.

        Applies source highlighting to hypothesis output that is not related to
        `hypothesis.errors.MultipleFailures`.

        Attributes
        ----------
        item : pytest.Item
            Object representing the currently active test
        """
        if not hasattr(item, "hypothesis_report_information"):
            return

        reporter = self.config.pluginmanager.get_plugin("terminalreporter")

        report = [x for i in item.hypothesis_report_information for x in i.split("\n")]
        report = [self._reduce_path_strings(i) for i in report]
        report = [
            reporter._tw._highlight(i).rstrip("\n") if not i.lstrip().startswith("\x1b") else i
            for i in report
        ]

        item.hypothesis_report_information = report

    def pytest_terminal_summary(self, terminalreporter):
        """
        Add a section to terminal summary reporting.

        * When the `--disable-warnings` flag is active, removes all raised warnings
          prior to outputting the final console report.

        * When `--coverage` is active, outputs the result to stdout and saves the
          final report json.

        Arguments
        ---------
        terminalreporter : `_pytest.terminal.TerminalReporter`
            The internal terminal reporter object
        """
        if self.config.getoption("--disable-warnings") and "warnings" in terminalreporter.stats:
            del terminalreporter.stats["warnings"]

        if CONFIG.argv["coverage"]:
            terminalreporter.section("Coverage")

            # output coverage report to console
            coverage_eval = coverage.get_merged_coverage_eval()
            for line in output._build_coverage_output(coverage_eval):
                terminalreporter.write_line(line)

            # save coverage report as `reports/coverage.json`
            output._save_coverage_report(
                self.project._build,
                coverage_eval,
                self.project_path.joinpath(self.project._structure["reports"]),
            )

    def pytest_unconfigure(self):
        """
        Called before test process is exited.

        Closes all active projects.
        """
        for project in brownie.project.get_loaded_projects():
            project.close(raises=False)

    def pytest_keyboard_interrupt(self):
        CONFIG.argv["interrupt"] = True



#!/usr/bin/python3

from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from eth_abi.grammar import ABIType, TupleType, parse

from .datatypes import EthAddress, HexString, ReturnValue
from .main import to_bool, to_decimal, to_int, to_string, to_uint
from .utils import get_type_strings


def format_input(abi: Dict, inputs: Union[List, Tuple]) -> List:
    # Format contract inputs based on ABI types
    if len(inputs) and not len(abi["inputs"]):
        raise TypeError(f"{abi['name']} requires no arguments")
    abi_types = _get_abi_types(abi["inputs"])
    try:
        return _format_tuple(abi_types, inputs)
    except Exception as e:
        raise type(e)(f"{abi['name']} {e}") from None


def format_output(abi: Dict, outputs: Union[List, Tuple]) -> ReturnValue:
    # Format contract outputs based on ABI types
    abi_types = _get_abi_types(abi["outputs"])
    result = _format_tuple(abi_types, outputs)
    return ReturnValue(result, abi["outputs"])


def format_event(event: Dict) -> Any:
    # Format event data based on ABI types
    if not event["decoded"]:
        topics = [
            {"type": "bytes32", "name": f"topic{c}", "value": i}
            for c, i in enumerate(event.get("topics", []), start=1)
        ]
        event["data"] = topics + [
            {"type": "bytes", "name": "data", "value": _format_single("bytes", event["data"])}
        ]
        if "anonymous" in event:
            event["name"] = "(anonymous)"
        else:
            event["name"] = "(unknown)"
        return event

    for e in [i for i in event["data"] if not i["decoded"]]:
        e["type"] = "bytes32"
        e["name"] += " (indexed)"
    abi_types = _get_abi_types(event["data"])
    values = ReturnValue(
        _format_tuple(abi_types, [i["value"] for i in event["data"]]), event["data"]
    )
    for i in range(len(event["data"])):
        event["data"][i]["value"] = values[i]
    return event


def _format_tuple(abi_types: Sequence[ABIType], values: Union[List, Tuple]) -> List:
    result = []
    _check_array(values, len(abi_types))
    for type_, value in zip(abi_types, values):
        try:
            if type_.is_array:
                result.append(_format_array(type_, value))
            elif isinstance(type_, TupleType):
                result.append(_format_tuple(type_.components, value))
            else:
                result.append(_format_single(type_.to_type_str(), value))
        except Exception as e:
            raise type(e)(f"'{value}' - {e}") from None
    return result


def _format_array(abi_type: ABIType, values: Union[List, Tuple]) -> List:
    _check_array(values, None if not len(abi_type.arrlist[-1]) else abi_type.arrlist[-1][0])
    item_type = abi_type.item_type
    if item_type.is_array:
        return [_format_array(item_type, i) for i in values]
    elif isinstance(item_type, TupleType):
        return [_format_tuple(item_type.components, i) for i in values]
    return [_format_single(item_type.to_type_str(), i) for i in values]


def _format_single(type_str: str, value: Any) -> Any:
    # Apply standard formatting to a single value
    if "uint" in type_str:
        return to_uint(value, type_str)
    elif "int" in type_str:
        return to_int(value, type_str)
    elif type_str == "fixed168x10":
        return to_decimal(value)
    elif type_str == "bool":
        return to_bool(value)
    elif type_str == "address":
        return EthAddress(value)
    elif "byte" in type_str:
        return HexString(value, type_str)
    elif "string" in type_str:
        return to_string(value)
    raise TypeError(f"Unknown type: {type_str}")


def _check_array(values: Union[List, Tuple], length: Optional[int]) -> None:
    if not isinstance(values, (list, tuple)):
        raise TypeError(f"Expected list or tuple, got {type(values).__name__}")
    if length is not None and len(values) != length:
        raise ValueError(f"Sequence has incorrect length, expected {length} but got {len(values)}")


def _get_abi_types(abi_params: List) -> Sequence[ABIType]:
    type_str = f"({','.join(get_type_strings(abi_params))})"
    tuple_type = parse(type_str)
    return tuple_type.components



#!/usr/bin/python3

from brownie import project
from brownie._gui import Gui
from brownie.utils.docopt import docopt

__doc__ = """Usage: brownie gui

Options:
  --help -h                  Display this message

Opens the brownie GUI. Basic functionality is as follows:

 * Selecting an opcode will highlight the associated source code.
 * Highlighting a section of the source will jump to the most relevent opcode,
   if possible.
 * Opcodes with a darkened background have no associated source code.
 * Type a pc number to jump to that opcode.
 * Right click an opcode to toggle highlighting on all opcodes of the same type.
 * Press J to toggle highlighting on JUMP, JUMPI and JUMPDEST opcodes.
 * Press R to toggle highlighting on all REVERT opcodes.
 * Select a section of source code and press S to enter scope mode. The
   instructions will be filtered to only display opcodes related to the relevent
   code. Press A to disable and see all opcodes again.
 * Press C to toggle unit test coverage visualization. This will only work if
   you have already run brownie coverage on your project. The coverage results
   are shown via different colors of text highlight."""


def main():
    docopt(__doc__)
    project.load()
    print("Loading Brownie GUI...")
    Gui().mainloop()
    print("GUI was terminated.")



#!/usr/bin/python3

import json
import shutil
import sys

import yaml

from brownie import accounts, network
from brownie._config import BROWNIE_FOLDER, _get_data_folder
from brownie.exceptions import ProjectNotFound, UnknownAccount
from brownie.project import check_for_project, ethpm
from brownie.utils import color, notify
from brownie.utils.docopt import docopt

__doc__ = """Usage: brownie ethpm <command> [<arguments> ...] [options]

Commands:
  list                             List installed packages
  install <uri> [overwrite=False]  Install a new package
  unlink <name>                    Unlink an installed package
  remove <name>                    Remove an installed package
  create [filename]                Generate a manifest
  release <registry> <account>     Generate a manifest and publish to an ethPM registry
  all                              List all locally available packages

Options:
  --help -h                        Display this message

ethPM is a decentralized package manager used to distribute EVM smart contracts
and projects. See https://eth-brownie.readthedocs.io/en/stable/ethpm.html for more
information on how to use it within Brownie.
"""


def main():
    args = docopt(__doc__)
    try:
        fn = getattr(sys.modules[__name__], f"_{args['<command>']}")
    except AttributeError:
        print("Invalid command. Try brownie ethpm --help")
        return
    project_path = check_for_project(".")
    if project_path is None:
        raise ProjectNotFound
    if not project_path.joinpath("ethpm-config.yaml").exists():
        shutil.copy(
            BROWNIE_FOLDER.joinpath("data/ethpm-config.yaml"),
            project_path.joinpath("ethpm-config.yaml"),
        )
    try:
        fn(project_path, *args["<arguments>"])
    except TypeError:
        print(f"Invalid arguments for command '{args['<command>']}'. Try brownie ethpm --help")
        return


def _list(project_path):
    installed, modified = ethpm.get_installed_packages(project_path)
    package_list = sorted(installed + modified)
    if not package_list:
        print("No packages are currently installed in this project.")
        return
    if modified:
        notify(
            "WARNING",
            f"One or more files in {len(modified)} packages have been modified since installation.",
        )
        print("Unlink or reinstall them to silence this warning.")
        print(f"Modified packages name are highlighted in {color('bright blue')}blue{color}.\n")
    print(f"Found {color('bright magenta')}{len(package_list)}{color} installed packages:")
    for name in package_list:
        u = "\u2514" if name == package_list[-1] else "\u251c"
        c = color("bright blue") if name in modified else color("bright white")
        print(
            f" {color('bright black')}{u}\u2500{c}{name[0]}{color}@"
            f"{color('bright white')}{name[1]}{color}"
        )


def _install(project_path, uri, replace=False):
    if replace:
        if replace.lower() not in ("true", "false"):
            raise ValueError("Invalid command for 'overwrite', must be True or False")
        replace = eval(replace.capitalize())
    print(f'Attempting to install package at "{color("bright magenta")}{uri}{color}"...')
    name = ethpm.install_package(project_path, uri, replace)
    notify("SUCCESS", f'The "{color("bright magenta")}{name}{color}" package was installed.')


def _unlink(project_path, name):
    if ethpm.remove_package(project_path, name, False):
        notify("SUCCESS", f'The "{color("bright magenta")}{name}{color}" package was unlinked.')
        return
    notify("ERROR", f'"{color("bright magenta")}{name}{color}" is not installed in this project.')


def _remove(project_path, name):
    if ethpm.remove_package(project_path, name, True):
        notify("SUCCESS", f'The "{color("bright magenta")}{name}{color}" package was removed.')
        return
    notify("ERROR", f'"{color("bright magenta")}{name}{color}" is not installed in this project.')


def _create(project_path, manifest_pathstr="manifest.json"):
    print("Generating a manifest based on configuration settings in ethpm-config.yaml...")
    with project_path.joinpath("ethpm-config.yaml").open() as fp:
        project_config = yaml.safe_load(fp)
    try:
        manifest = ethpm.create_manifest(project_path, project_config)[0]
    except Exception as e:
        notify("ERROR", f"{type(e).__name__}: {e}")
        print("Ensure that all package configuration settings are correct in ethpm-config.yaml")
        return
    with project_path.joinpath(manifest_pathstr).open("w") as fp:
        json.dump(manifest, fp, sort_keys=True, indent=2)
    notify(
        "SUCCESS",
        f'Generated manifest saved as "{color("bright magenta")}{manifest_pathstr}{color}"',
    )


def _release(project_path, registry_address, sender):
    network.connect("mainnet")
    with project_path.joinpath("ethpm-config.yaml").open() as fp:
        project_config = yaml.safe_load(fp)
    print("Generating manifest and pinning assets to IPFS...")
    manifest, uri = ethpm.create_manifest(project_path, project_config, True, False)
    if sender in accounts:
        account = accounts.at(sender)
    else:
        try:
            account = accounts.load(sender)
        except FileNotFoundError:
            raise UnknownAccount(f"Unknown account '{sender}'")
    name = f'{manifest["package_name"]}@{manifest["version"]}'
    print(f'Releasing {name} on "{registry_address}"...')
    try:
        tx = ethpm.release_package(
            registry_address, account, manifest["package_name"], manifest["version"], uri
        )
        if tx.status == 1:
            notify("SUCCESS", f"{name} has been released!")
            print(f"\nURI: {color('bright magenta')}ethpm://{registry_address}:1/{name}{color}")
            return
    except Exception:
        pass
    notify("ERROR", f'Transaction reverted when releasing {name} on "{registry_address}"')


def _all(project_path):
    ethpm_folder = _get_data_folder().joinpath("ethpm")
    print(
        f"Visit {color('bright magenta')}https://explorer.ethpm.com/{color}"
        " for a list of ethPM registries and packages.\n"
    )
    if not list(ethpm_folder.glob("*")):
        print("No local packages are currently available.")
    for path in sorted(ethpm_folder.glob("*")):
        package_list = sorted(path.glob("*"))
        if not package_list:
            path.unlink()
            continue
        print(f"{color('bright magenta')}ethpm://{path.name}{color}")
        for package_path in package_list:
            u = "\u2514" if package_path == package_list[-1] else "\u251c"
            versions = sorted(package_path.glob("*.json"))
            if len(versions) == 1:
                print(
                    f" {color('bright black')}{u}\u2500{color('bright white')}{package_path.stem}"
                    f"{color}@{color('bright white')}{versions[0].stem}{color}"
                )
                continue
            print(
                f" {color('bright black')}{u}\u2500{color('bright white')}"
                f"{package_path.stem}{color}"
            )
            for v in versions:
                u = "\u2514" if v == versions[-1] else "\u251c"
                print(f"   {color('bright black')}{u}\u2500{color('bright white')}{v.stem}{color}")



#!/usr/bin/python3

from typing import Dict, ItemsView, List, Optional, Tuple, Union

from .sources import Sources, highlight_source

INTERFACE_KEYS = ["abi", "contractName", "sha1", "type"]

DEPLOYMENT_KEYS = [
    "abi",
    "ast",
    "bytecode",
    "compiler",
    "contractName",
    "deployedBytecode",
    "deployedSourceMap",
    "language",
    "natspec",
    "opcodes",
    "pcMap",
    "sourceMap",
    "type",
]

BUILD_KEYS = [
    "allSourcePaths",
    "bytecodeSha1",
    "coverageMap",
    "dependencies",
    "offset",
    "sha1",
    "source",
    "sourcePath",
] + DEPLOYMENT_KEYS

_revert_map: Dict = {}


class Build:

    """Methods for accessing and manipulating a project's contract build data."""

    def __init__(self, sources: Sources) -> None:
        self._sources = sources
        self._contracts: Dict = {}
        self._interfaces: Dict = {}

    def _add_contract(self, build_json: Dict) -> None:
        contract_name = build_json["contractName"]
        if contract_name in self._contracts and build_json["type"] == "interface":
            return
        self._contracts[contract_name] = build_json
        if "pcMap" not in build_json:
            # no pcMap means build artifact is for an interface
            return
        if "0" in build_json["pcMap"]:
            build_json["pcMap"] = dict((int(k), v) for k, v in build_json["pcMap"].items())
        self._generate_revert_map(
            build_json["pcMap"], build_json["allSourcePaths"], build_json["language"]
        )

    def _add_interface(self, build_json: Dict) -> None:
        contract_name = build_json["contractName"]
        self._interfaces[contract_name] = build_json

    def _generate_revert_map(self, pcMap: Dict, source_map: Dict, language: str) -> None:
        # Adds a contract's dev revert strings to the revert map and it's pcMap
        marker = "//" if language == "Solidity" else "#"
        for pc, data in (
            (k, v)
            for k, v in pcMap.items()
            if v["op"] in ("REVERT", "INVALID") or "jump_revert" in v
        ):
            if "path" not in data:
                continue
            path_str = source_map[data["path"]]

            if "dev" not in data:
                if "fn" not in data or "first_revert" in data:
                    _revert_map[pc] = False
                    continue
                try:
                    revert_str = self._sources.get(path_str)[data["offset"][1] :]
                    revert_str = revert_str[: revert_str.index("\n")]
                    revert_str = revert_str[revert_str.index(marker) + len(marker) :].strip()
                    if revert_str.startswith("dev:"):
                        data["dev"] = revert_str
                except (KeyError, ValueError):
                    pass

            msg = "" if data["op"] == "REVERT" else "invalid opcode"
            revert = (
                path_str,
                tuple(data["offset"]),
                data.get("fn", "<None>"),
                data.get("dev", msg),
                self._sources,
            )

            # do not compare the final tuple item in case the same project was loaded twice
            if pc not in _revert_map or (_revert_map[pc] and revert[:-1] == _revert_map[pc][:-1]):
                _revert_map[pc] = revert
                continue
            _revert_map[pc] = False

    def _remove_contract(self, contract_name: str) -> None:
        key = self._stem(contract_name)
        if key in self._contracts:
            del self._contracts[key]

    def _remove_interface(self, contract_name: str) -> None:
        key = self._stem(contract_name)
        if key in self._interfaces:
            del self._interfaces[key]

    def get(self, contract_name: str) -> Dict:
        """Returns build data for the given contract name."""
        return self._contracts[self._stem(contract_name)]

    def items(self, path: Optional[str] = None) -> Union[ItemsView, List]:
        """Provides an list of tuples as (key,value), similar to calling dict.items.
        If a path is given, only contracts derived from that source file are returned."""
        items = list(self._contracts.items()) + list(self._interfaces.items())
        if path is None:
            return items
        return [(k, v) for k, v in items if v.get("sourcePath") == path]

    def contains(self, contract_name: str) -> bool:
        """Checks if the contract name exists in the currently loaded build data."""
        return self._stem(contract_name) in list(self._contracts) + list(self._interfaces)

    def get_dependents(self, contract_name: str) -> List:
        """Returns a list of contract names that inherit from or link to the given
        contract. Used by the compiler when determining which contracts to recompile
        based on a changed source file."""
        return [k for k, v in self._contracts.items() if contract_name in v.get("dependencies", [])]

    def _stem(self, contract_name: str) -> str:
        return contract_name.replace(".json", "")


def _get_dev_revert(pc: int) -> Optional[str]:
    # Given the program counter from a stack trace that caused a transaction
    # to revert, returns the commented dev string (if any)
    if pc not in _revert_map or _revert_map[pc] is False:
        return None
    return _revert_map[pc][3]


def _get_error_source_from_pc(pc: int, pad: int = 3) -> Tuple:
    # Given the program counter from a stack trace that caused a transaction
    # to revert, returns the highlighted relevent source code and the method name.
    if pc not in _revert_map or _revert_map[pc] is False:
        return (None,) * 4
    revert = _revert_map[pc]
    source = revert[4].get(revert[0])
    highlight, linenos = highlight_source(source, revert[1], pad=pad)  # type: ignore
    return highlight, linenos, revert[0], revert[2]



#!/usr/bin/python3

import sys

import pytest

import brownie
from brownie._config import CONFIG, _get_data_folder

from .stateful import _BrownieStateMachine, state_machine


def _generate_fixture(container):
    def _fixture():
        yield container

    _fixture.__doc__ = f"Provides access to Brownie ContractContainer object '{container._name}'"
    return pytest.fixture(scope="session")(_fixture)


class PytestBrownieFixtures:
    def __init__(self, config, project):
        self.config = config
        self._interface = project.interface
        for container in project:
            setattr(self, container._name, _generate_fixture(container))

    # fixtures
    @pytest.fixture(scope="module")
    def module_isolation(self):
        """
        Resets the test environment before and after a test module runs. This ensures
        a clean environment for the module, and that it's results will not affect
        subsequent tests.
        """
        brownie.chain.reset()
        yield
        if not CONFIG.argv["interrupt"]:
            brownie.chain.reset()

    @pytest.fixture
    def fn_isolation(self, module_isolation):
        """
        Performs the actions of module_isolation, takes a snapshot after all module
        scoped fixtures have run, and reverts to this snapshot at the start of each test.

        Used to ensure that each test in a module begins with an identical environment.
        """
        brownie.chain.snapshot()
        yield
        if not CONFIG.argv["interrupt"]:
            brownie.chain.revert()

    @pytest.fixture(scope="session")
    def accounts(self):
        """Yields an Accounts container for the active project, used to access local accounts."""
        yield brownie.accounts

    @pytest.fixture(scope="session")
    def a(self):
        """Short form of the accounts fixture."""
        yield brownie.accounts

    @pytest.fixture(scope="session")
    def chain(self):
        """Yields a Chain object, used for interacting with the blockchain."""
        yield brownie.chain

    @pytest.fixture(scope="session")
    def Contract(self):
        """Yields the Contract class, used to interact with deployments outside of a project."""
        yield brownie.Contract

    @pytest.fixture(scope="session")
    def history(self):
        """Yields a TxHistory container for the active project, used to access transaction data."""
        yield brownie.history

    @pytest.fixture(scope="session")
    def interface(self):
        yield self._interface

    @pytest.fixture(scope="session")
    def rpc(self):
        """Yields an Rpc object, used for interacting with the local RPC client."""
        yield brownie.rpc

    @pytest.fixture(scope="session")
    def web3(self):
        """Yields an instantiated Web3 object, connected to the active network."""
        yield brownie.web3

    @pytest.fixture(scope="session")
    def pm(self):
        """
        Yields a function for accessing installed packages.
        """
        _open_projects = {}

        def package_loader(project_id):
            if project_id not in _open_projects:
                path = _get_data_folder().joinpath(f"packages/{project_id}")
                _open_projects[project_id] = brownie.project.load(path, project_id)

            return _open_projects[project_id]

        yield package_loader

    @pytest.fixture
    def no_call_coverage(self):
        """
        Prevents coverage evaluation on contract calls during this test. Useful for speeding
        up tests that contain many repetetive calls.
        """
        CONFIG.argv["always_transact"] = False
        yield
        CONFIG.argv["always_transact"] = CONFIG.argv["coverage"]

    @pytest.fixture(scope="session")
    def skip_coverage(self):
        """Skips a test when coverage evaluation is active."""
        # implemented in pytest_collection_modifyitems
        pass

    @pytest.fixture
    def state_machine(self):
        """Yields a rule-based state machine factory method."""

        if self.config.getoption("capture") != "no":
            # allows the state machine to disable pytest capturing
            capman = self.config.pluginmanager.get_plugin("capturemanager")
            _BrownieStateMachine._capman = capman

            # for posix systems we disable the cursor to make the progress spinner prettier
            if sys.platform != "win32":
                with capman.global_and_fixture_disabled():
                    sys.stdout.write("\033[?25l")
                    sys.stdout.flush()

        yield state_machine

        if self.config.getoption("capture") != "no" and sys.platform != "win32":
            # re-enable the cursor
            with capman.global_and_fixture_disabled():
                sys.stdout.write("\033[?25h")
                sys.stdout.flush()



#!/usr/bin/python3

import sys
from collections import deque
from inspect import getmembers
from types import FunctionType
from typing import Any, Optional

from hypothesis import settings as hp_settings
from hypothesis import stateful as sf
from hypothesis.strategies import SearchStrategy

import brownie
from brownie.utils import color

sf.__tracebackhide__ = True

marker = deque("-/|\\-/|\\")


class _BrownieStateMachine:

    _failed = False

    def __init__(self) -> None:
        brownie.chain.revert()
        sf.RuleBasedStateMachine.__init__(self)

        # pytest capturemanager plugin, added when accessed via the state_manager fixture
        capman = getattr(self, "_capman", None)
        if capman:
            with capman.global_and_fixture_disabled():
                c = color("red" if self._failed else "yellow")
                sys.stdout.write(f"{c}{marker[0]}\033[1D")
                sys.stdout.flush()
            marker.rotate(1)

        if hasattr(self, "setup"):
            self.setup()  # type: ignore

    def execute_step(self, step):
        try:
            super().execute_step(step)
        except Exception:
            type(self)._failed = True
            raise

    def check_invariants(self):
        try:
            super().check_invariants()
        except Exception:
            type(self)._failed = True
            raise


def _member_filter(member: tuple) -> bool:
    attr, fn = member
    return (
        type(fn) is FunctionType
        and not hasattr(sf.RuleBasedStateMachine, attr)
        and not next((i for i in fn.__dict__.keys() if i.startswith("hypothesis_stateful")), False)
    )


def _attr_filter(attr: str, pattern: str) -> bool:
    return attr == pattern or attr.startswith(f"{pattern}_")


def _generate_state_machine(rules_object: type) -> type:

    bases = (_BrownieStateMachine, rules_object, sf.RuleBasedStateMachine)
    machine = type("BrownieStateMachine", bases, {})
    strategies = {k: v for k, v in getmembers(rules_object) if isinstance(v, SearchStrategy)}

    for attr, fn in filter(_member_filter, getmembers(machine)):
        varnames = [[i] for i in fn.__code__.co_varnames[1 : fn.__code__.co_argcount]]
        if fn.__defaults__:
            for i in range(-1, -1 - len(fn.__defaults__), -1):
                varnames[i].append(fn.__defaults__[i])

        if _attr_filter(attr, "initialize"):
            wrapped = sf.initialize(**{key[0]: strategies[key[-1]] for key in varnames})
            setattr(machine, attr, wrapped(fn))
        elif _attr_filter(attr, "invariant"):
            setattr(machine, attr, sf.invariant()(fn))
        elif _attr_filter(attr, "rule"):
            wrapped = sf.rule(**{key[0]: strategies[key[-1]] for key in varnames})
            setattr(machine, attr, wrapped(fn))

    return machine


def state_machine(
    rules_object: type, *args: Any, settings: Optional[dict] = None, **kwargs: Any
) -> None:

    machine = _generate_state_machine(rules_object)
    if hasattr(rules_object, "__init__"):
        # __init__ is treated as a class method
        rules_object.__init__(machine, *args, **kwargs)  # type: ignore
    brownie.chain.snapshot()

    try:
        sf.run_state_machine_as_test(lambda: machine(), settings=hp_settings(**settings or {}))
    finally:
        if hasattr(machine, "teardown_final"):
            # teardown_final is also a class method
            machine.teardown_final(machine)  # type: ignore



#!/usr/bin/python3
from typing import Any, Dict, Tuple


class _Singleton(type):

    _instances: Dict = {}

    def __call__(cls, *args: Tuple, **kwargs: Dict) -> Any:
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



#!/usr/bin/python3

import json
import warnings
from collections import OrderedDict
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Sequence, Tuple, Union, ValuesView

import eth_event
from eth_event import EventError

from brownie._config import _get_data_folder
from brownie.convert.normalize import format_event
from brownie.exceptions import EventLookupError


class EventDict:
    """
    Dict/list hybrid container, base class for all events fired in a transaction.
    """

    def __init__(self, events: List) -> None:
        """Instantiates the class.

        Args:
            events: event data as supplied by eth_event.decode_logs or eth_event.decode_trace"""
        self._ordered = [
            _EventItem(
                i["name"],
                i["address"],
                [OrderedDict((x["name"], x["value"]) for x in i["data"])],
                (pos,),
            )
            for pos, i in enumerate(events)
        ]

        self._dict: Dict = OrderedDict()
        for event in self._ordered:
            if event.name not in self._dict:
                events = [i for i in self._ordered if i.name == event.name]
                self._dict[event.name] = _EventItem(
                    event.name, None, events, tuple(i.pos[0] for i in events)
                )

    def __repr__(self) -> str:
        return str(self)

    def __bool__(self) -> bool:
        return bool(self._ordered)

    def __contains__(self, name: str) -> bool:
        """returns True if an event fired with the given name."""
        return name in [i.name for i in self._ordered]

    def __getitem__(self, key: Union[str, int]) -> "_EventItem":
        """if key is int: returns the n'th event that was fired
        if key is str: returns a _EventItem dict of all events where name == key"""
        if not isinstance(key, (int, str)):
            raise TypeError(f"Invalid key type '{type(key)}' - can only use strings or integers")
        if isinstance(key, int):
            try:
                return self._ordered[key]
            except IndexError:
                raise EventLookupError(
                    f"Index out of range - only {len(self._ordered)} events fired"
                )
        if key in self._dict:
            return self._dict[key]
        raise EventLookupError(f"Event '{key}' did not fire.")

    def __iter__(self) -> Iterator:
        return iter(self._ordered)

    def __len__(self) -> int:
        """returns the number of events that fired."""
        return len(self._ordered)

    def __str__(self) -> str:
        return str(dict((k, [i[0] for i in v._ordered]) for k, v in self._dict.items()))

    def count(self, name: str) -> int:
        """EventDict.count(name) -> integer -- return number of occurrences of name"""
        return len([i.name for i in self._ordered if i.name == name])

    def items(self) -> List:
        """EventDict.items() -> a list object providing a view on EventDict's items"""
        return list(self._dict.items())

    def keys(self) -> List:
        """EventDict.keys() -> a list object providing a view on EventDict's keys"""
        return list(self._dict.keys())

    def values(self) -> ValuesView:
        """EventDict.values() -> a list object providing a view on EventDict's values"""
        return self._dict.values()


class _EventItem:
    """
    Dict/list hybrid container, represents one or more events with the same name
    that were fired in a transaction.

    Attributes
    ----------
    name : str
        Name of the event.
    address : str
        Address where this event fired. When the object represents more than one event,
        this value is set to `None`.
    pos : tuple
        Tuple of indexes where this event fired.
    """

    def __init__(self, name: str, address: Optional[str], event_data: List, pos: Tuple) -> None:
        self.name = name
        self.address = address
        self._ordered = event_data
        self.pos = pos

    def __getitem__(self, key: Union[int, str]) -> List:
        """if key is int: returns the n'th event that was fired with this name
        if key is str: returns the value of data field 'key' from the 1st event
        within the container """
        if not isinstance(key, (int, str)):
            raise TypeError(f"Invalid key type '{type(key)}' - can only use strings or integers")
        if isinstance(key, int):
            try:
                return self._ordered[key]
            except IndexError:
                raise EventLookupError(
                    f"Index out of range - only {len(self._ordered)} '{self.name}' events fired"
                )
        if key in self._ordered[0]:
            return self._ordered[0][key]
        if f"{key} (indexed)" in self._ordered[0]:
            return self._ordered[0][f"{key} (indexed)"]
        valid_keys = ", ".join(self.keys())
        raise EventLookupError(
            f"Unknown key '{key}' - the '{self.name}' event includes these keys: {valid_keys}"
        )

    def __contains__(self, name: str) -> bool:
        """returns True if this event contains a value with the given name."""
        return name in self._ordered[0]

    def __len__(self) -> int:
        """returns the number of events held in this container."""
        return len(self._ordered)

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        if len(self._ordered) == 1:
            return str(self._ordered[0])
        return str([i[0] for i in self._ordered])

    def __iter__(self) -> Iterator:
        return iter(self._ordered)

    def __eq__(self, other: object) -> bool:
        if len(self._ordered) == 1:
            return other == self._ordered[0]
        return other == self._ordered

    def items(self) -> List:
        """_EventItem.items() -> a list object providing a view on _EventItem[0]'s items"""
        return [(i, self[i]) for i in self.keys()]

    def keys(self) -> List:
        """_EventItem.keys() -> a list object providing a view on _EventItem[0]'s keys"""
        return [i.replace(" (indexed)", "") for i in self._ordered[0].keys()]

    def values(self) -> List:
        """_EventItem.values() -> a list object providing a view on _EventItem[0]'s values"""
        return list(self._ordered[0].values())


def __get_path() -> Path:
    return _get_data_folder().joinpath("topics.json")


def _get_topics(abi: List) -> Dict:
    topic_map = eth_event.get_topic_map(abi)

    updated_topics = _topics.copy()

    for key, value in topic_map.items():
        if key not in updated_topics:
            # new event topic
            updated_topics[key] = value
        elif value == updated_topics[key]:
            # existing event topic, nothing has changed
            continue
        elif not next((i for i in updated_topics[key]["inputs"] if i["indexed"]), False):
            # existing topic, but the old abi has no indexed events - keep the new one
            updated_topics[key] = value

    if updated_topics != _topics:
        _topics.update(updated_topics)
        with __get_path().open("w") as fp:
            json.dump(updated_topics, fp, sort_keys=True, indent=2)

    return {v["name"]: k for k, v in topic_map.items()}


def _add_deployment_topics(address: str, abi: List) -> None:
    _deployment_topics[address] = eth_event.get_topic_map(abi)


def _decode_logs(logs: List) -> Union["EventDict", List[None]]:
    if not logs:
        return []

    idx = 0
    events: List = []
    while True:
        address = logs[idx]["address"]
        try:
            new_idx = logs.index(next(i for i in logs[idx:] if i["address"] != address))
            log_slice = logs[idx:new_idx]
            idx = new_idx
        except StopIteration:
            log_slice = logs[idx:]

        topics_map = _deployment_topics.get(address, _topics)
        for item in log_slice:
            try:
                events.extend(eth_event.decode_logs([item], topics_map, allow_undecoded=True))
            except EventError as exc:
                warnings.warn(str(exc))

        if log_slice[-1] == logs[-1]:
            break

    events = [format_event(i) for i in events]
    return EventDict(events)


def _decode_trace(trace: Sequence, initial_address: str) -> Union["EventDict", List[None]]:
    if not trace:
        return []

    events = eth_event.decode_traceTransaction(
        trace, _topics, allow_undecoded=True, initial_address=initial_address
    )
    events = [format_event(i) for i in events]
    return EventDict(events)


# dictionary of event topic ABIs specific to a single contract deployment
_deployment_topics: Dict = {}

# general event topic ABIs for decoding events on unknown contracts
_topics: Dict = {}

try:
    with __get_path().open() as fp:
        _topics = json.load(fp)
except (FileNotFoundError, json.decoder.JSONDecodeError):
    pass



#!/usr/bin/python3

import json
from copy import deepcopy
from hashlib import sha1
from pathlib import Path
from typing import Dict, Optional

from eth_utils import remove_0x_prefix
from semantic_version import Version

from brownie._config import _get_data_folder
from brownie.exceptions import UnsupportedLanguage
from brownie.project import sources
from brownie.project.compiler.solidity import (  # NOQA: F401
    find_best_solc_version,
    find_solc_versions,
    install_solc,
    set_solc_version,
)
from brownie.project.compiler.utils import merge_natspec
from brownie.project.compiler.vyper import find_vyper_versions, set_vyper_version
from brownie.utils import notify

from . import solidity, vyper

STANDARD_JSON: Dict = {
    "language": None,
    "sources": {},
    "settings": {
        "outputSelection": {
            "*": {
                "*": ["abi", "devdoc", "evm.bytecode", "evm.deployedBytecode", "userdoc"],
                "": ["ast"],
            }
        },
        "evmVersion": None,
        "remappings": [],
    },
}
EVM_SOLC_VERSIONS = [
    ("istanbul", Version("0.5.13")),
    ("petersburg", Version("0.5.5")),
    ("byzantium", Version("0.4.0")),
]


def compile_and_format(
    contract_sources: Dict[str, str],
    solc_version: Optional[str] = None,
    vyper_version: Optional[str] = None,
    optimize: bool = True,
    runs: int = 200,
    evm_version: Optional[str] = None,
    silent: bool = True,
    allow_paths: Optional[str] = None,
    interface_sources: Optional[Dict[str, str]] = None,
    remappings: Optional[list] = None,
    optimizer: Optional[Dict] = None,
) -> Dict:
    """Compiles contracts and returns build data.

    Args:
        contract_sources: a dictionary in the form of {'path': "source code"}
        solc_version: solc version to compile with (use None to set via pragmas)
        optimize: (deprecated) enable solc optimizer
        runs: (deprecated) optimizer runs
        evm_version: evm version to compile for
        silent: verbose reporting
        allow_paths: compiler allowed filesystem import path
        interface_sources: dictionary of interfaces as {'path': "source code"}
        remappings: list of solidity path remappings
        optimizer: dictionary of solidity optimizer settings

    Returns:
        build data dict
    """
    if not contract_sources:
        return {}
    if interface_sources is None:
        interface_sources = {}

    if [i for i in contract_sources if Path(i).suffix not in (".sol", ".vy")]:
        raise UnsupportedLanguage("Source suffixes must be one of ('.sol', '.vy')")
    if [i for i in interface_sources if Path(i).suffix not in (".sol", ".vy", ".json")]:
        raise UnsupportedLanguage("Interface suffixes must be one of ('.sol', '.vy', '.json')")

    build_json: Dict = {}
    compiler_targets = {}

    vyper_sources = {k: v for k, v in contract_sources.items() if Path(k).suffix == ".vy"}
    if vyper_sources:
        # TODO add `vyper_version` input arg to manually specify, support in config file
        if vyper_version is None:
            compiler_targets.update(
                find_vyper_versions(vyper_sources, install_needed=True, silent=silent)
            )
        else:
            compiler_targets[vyper_version] = list(vyper_sources)
    solc_sources = {k: v for k, v in contract_sources.items() if Path(k).suffix == ".sol"}
    if solc_sources:
        if solc_version is None:
            compiler_targets.update(
                find_solc_versions(solc_sources, install_needed=True, silent=silent)
            )
        else:
            compiler_targets[solc_version] = list(solc_sources)

        if optimizer is None:
            optimizer = {"enabled": optimize, "runs": runs if optimize else 0}

    for version, path_list in compiler_targets.items():
        compiler_data: Dict = {}
        if path_list[0].endswith(".vy"):
            set_vyper_version(version)
            language = "Vyper"
            compiler_data["version"] = str(vyper.get_version())
            interfaces = {k: v for k, v in interface_sources.items() if Path(k).suffix != ".sol"}
        else:
            set_solc_version(version)
            language = "Solidity"
            compiler_data["version"] = str(solidity.get_version())
            interfaces = {
                k: v
                for k, v in interface_sources.items()
                if Path(k).suffix == ".sol" and Version(version) in sources.get_pragma_spec(v, k)
            }

        to_compile = {k: v for k, v in contract_sources.items() if k in path_list}

        input_json = generate_input_json(
            to_compile,
            evm_version=evm_version,
            language=language,
            interface_sources=interfaces,
            remappings=remappings,
            optimizer=optimizer,
        )

        output_json = compile_from_input_json(input_json, silent, allow_paths)

        output_json["contracts"] = {
            k: v for k, v in output_json["contracts"].items() if k in path_list
        }

        build_json.update(generate_build_json(input_json, output_json, compiler_data, silent))
    return build_json


def generate_input_json(
    contract_sources: Dict[str, str],
    optimize: bool = True,
    runs: int = 200,
    evm_version: Optional[str] = None,
    language: str = "Solidity",
    interface_sources: Optional[Dict[str, str]] = None,
    remappings: Optional[list] = None,
    optimizer: Optional[Dict] = None,
) -> Dict:

    """Formats contracts to the standard solc input json.

    Args:
        contract_sources: a dictionary in the form of {path: 'source code'}
        optimize: (deprecated) enable solc optimizer
        runs: (deprecated) optimizer runs
        evm_version: evm version to compile for
        language: source language (Solidity or Vyper)
        interface_sources: dictionary of interfaces as {'path': "source code"}
        remappings: list of solidity path remappings
        optimizer: dictionary of solidity optimizer settings

    Returns: dict
    """

    if language not in ("Solidity", "Vyper"):
        raise UnsupportedLanguage(f"{language}")

    if optimizer is None:
        optimizer = {"enabled": optimize, "runs": runs if optimize else 0}

    if evm_version is None:
        if language == "Solidity":
            evm_version = next(i[0] for i in EVM_SOLC_VERSIONS if solidity.get_version() >= i[1])
        else:
            evm_version = "istanbul"

    input_json: Dict = deepcopy(STANDARD_JSON)
    input_json["language"] = language
    input_json["settings"]["evmVersion"] = evm_version
    if language == "Solidity":
        input_json["settings"]["optimizer"] = optimizer
        input_json["settings"]["remappings"] = _get_solc_remappings(remappings)
    input_json["sources"] = _sources_dict(contract_sources, language)

    if interface_sources:
        if language == "Solidity":
            input_json["sources"].update(_sources_dict(interface_sources, language))
        else:
            input_json["interfaces"] = _sources_dict(interface_sources, language)

    return input_json


def _get_solc_remappings(remappings: Optional[list]) -> list:
    if remappings is None:
        remap_dict: Dict = {}
    elif isinstance(remappings, str):
        remap_dict = dict([remappings.split("=")])
    else:
        remap_dict = dict(i.split("=") for i in remappings)

    for path in _get_data_folder().joinpath("packages").iterdir():
        key = next((k for k, v in remap_dict.items() if v.startswith(path.name)), None)
        if key:
            remap_dict[key] = path.parent.joinpath(remap_dict[key]).as_posix()
        else:
            remap_dict[path.name] = path.as_posix()

    return [f"{k}={v}" for k, v in remap_dict.items()]


def _get_allow_paths(allow_paths: Optional[str], remappings: list) -> str:
    # generate the final allow_paths field based on path remappings
    path_list = [] if allow_paths is None else [allow_paths]

    remapping_paths = [i[i.index("=") + 1 :] for i in remappings]
    data_path = _get_data_folder().joinpath("packages").as_posix()
    remapping_paths = [i for i in remapping_paths if not i.startswith(data_path)]

    path_list = path_list + [data_path] + remapping_paths
    return ",".join(path_list)


def compile_from_input_json(
    input_json: Dict, silent: bool = True, allow_paths: Optional[str] = None
) -> Dict:

    """
    Compiles contracts from a standard input json.

    Args:
        input_json: solc input json
        silent: verbose reporting
        allow_paths: compiler allowed filesystem import path

    Returns: standard compiler output json
    """

    if input_json["language"] == "Vyper":
        return vyper.compile_from_input_json(input_json, silent, allow_paths)

    if input_json["language"] == "Solidity":
        allow_paths = _get_allow_paths(allow_paths, input_json["settings"]["remappings"])
        return solidity.compile_from_input_json(input_json, silent, allow_paths)

    raise UnsupportedLanguage(f"{input_json['language']}")


def generate_build_json(
    input_json: Dict, output_json: Dict, compiler_data: Optional[Dict] = None, silent: bool = True
) -> Dict:
    """Formats standard compiler output to the brownie build json.

    Args:
        input_json: solc input json used to compile
        output_json: output json returned by compiler
        compiler_data: additonal data to include under 'compiler' in build json
        silent: verbose reporting

    Returns: build json dict"""

    if input_json["language"] not in ("Solidity", "Vyper"):
        raise UnsupportedLanguage(f"{input_json['language']}")

    if not silent:
        print("Generating build data...")

    if compiler_data is None:
        compiler_data = {}
    compiler_data["evm_version"] = input_json["settings"]["evmVersion"]
    build_json: Dict = {}
    path_list = list(input_json["sources"])

    if input_json["language"] == "Solidity":
        compiler_data["optimizer"] = input_json["settings"]["optimizer"]
        source_nodes, statement_nodes, branch_nodes = solidity._get_nodes(output_json)

    for path_str, contract_name in [
        (k, v) for k in path_list for v in output_json["contracts"].get(k, {})
    ]:

        if not silent:
            print(f" - {contract_name}...")

        abi = output_json["contracts"][path_str][contract_name]["abi"]
        natspec = merge_natspec(
            output_json["contracts"][path_str][contract_name].get("devdoc", {}),
            output_json["contracts"][path_str][contract_name].get("userdoc", {}),
        )
        output_evm = output_json["contracts"][path_str][contract_name]["evm"]
        if contract_name in build_json and not output_evm["deployedBytecode"]["object"]:
            continue

        if input_json["language"] == "Solidity":
            contract_node = next(
                i[contract_name] for i in source_nodes if i.absolutePath == path_str
            )
            build_json[contract_name] = solidity._get_unique_build_json(
                output_evm,
                contract_node,
                statement_nodes,
                branch_nodes,
                next((True for i in abi if i["type"] == "fallback"), False),
            )

        else:
            if contract_name == "<stdin>":
                contract_name = "Vyper"
            build_json[contract_name] = vyper._get_unique_build_json(
                output_evm,
                path_str,
                contract_name,
                output_json["sources"][path_str]["ast"],
                (0, len(input_json["sources"][path_str]["content"])),
            )

        build_json[contract_name].update(
            {
                "abi": abi,
                "ast": output_json["sources"][path_str]["ast"],
                "compiler": compiler_data,
                "contractName": contract_name,
                "deployedBytecode": output_evm["deployedBytecode"]["object"],
                "deployedSourceMap": output_evm["deployedBytecode"]["sourceMap"],
                "language": input_json["language"],
                "natspec": natspec,
                "opcodes": output_evm["deployedBytecode"]["opcodes"],
                "sha1": sha1(input_json["sources"][path_str]["content"].encode()).hexdigest(),
                "source": input_json["sources"][path_str]["content"],
                "sourceMap": output_evm["bytecode"].get("sourceMap", ""),
                "sourcePath": path_str,
            }
        )
        size = len(remove_0x_prefix(output_evm["deployedBytecode"]["object"])) / 2  # type: ignore
        if size > 24577:
            notify(
                "WARNING",
                f"deployed size of {contract_name} is {size} bytes, exceeds EIP-170 limit of 24577",
            )

    if not silent:
        print("")

    return build_json


def _sources_dict(original: Dict, language: str) -> Dict:
    result: Dict = {}
    for key, value in original.items():
        if Path(key).suffix == ".json":
            if isinstance(value, str):
                value = json.loads(value)
            result[key] = {"abi": value}
        else:
            result[key] = {"content": value}
    return result


def get_abi(
    contract_sources: Dict[str, str],
    allow_paths: Optional[str] = None,
    remappings: Optional[list] = None,
    silent: bool = True,
) -> Dict:
    """
    Generate ABIs from contract interfaces.

    Arguments
    ---------
    contract_sources : dict
        a dictionary in the form of {'path': "source code"}
    allow_paths : str, optional
        Compiler allowed filesystem import path
    remappings : list, optional
        List of solidity path remappings
    silent : bool, optional
        Disable verbose reporting

    Returns
    -------
    dict
        Compiled ABIs in the format `{'contractName': [ABI]}`
    """

    final_output = {
        Path(k).stem: {
            "abi": json.loads(v),
            "contractName": Path(k).stem,
            "type": "interface",
            "sha1": sha1(v.encode()).hexdigest(),
        }
        for k, v in contract_sources.items()
        if Path(k).suffix == ".json"
    }

    for path, source in [(k, v) for k, v in contract_sources.items() if Path(k).suffix == ".vy"]:
        input_json = generate_input_json({path: source}, language="Vyper")
        input_json["settings"]["outputSelection"]["*"] = {"*": ["abi"]}
        try:
            output_json = compile_from_input_json(input_json, silent, allow_paths)
        except Exception:
            # vyper interfaces do not convert to ABIs
            # https://github.com/vyperlang/vyper/issues/1944
            continue
        name = Path(path).stem
        final_output[name] = {
            "abi": output_json["contracts"][path][name]["abi"],
            "contractName": name,
            "type": "interface",
            "sha1": sha1(contract_sources[path].encode()).hexdigest(),
        }

    solc_sources = {k: v for k, v in contract_sources.items() if Path(k).suffix == ".sol"}

    if solc_sources:

        compiler_targets = find_solc_versions(solc_sources, install_needed=True, silent=silent)

        for version, path_list in compiler_targets.items():
            to_compile = {k: v for k, v in contract_sources.items() if k in path_list}

            set_solc_version(version)
            input_json = generate_input_json(to_compile, language="Solidity", remappings=remappings)
            input_json["settings"]["outputSelection"]["*"] = {"*": ["abi"]}

            output_json = compile_from_input_json(input_json, silent, allow_paths)
            output_json = {k: v for k, v in output_json["contracts"].items() if k in path_list}

            final_output.update(
                {
                    name: {
                        "abi": data["abi"],
                        "contractName": name,
                        "type": "interface",
                        "sha1": sha1(contract_sources[path].encode()).hexdigest(),
                    }
                    for path, v in output_json.items()
                    for name, data in v.items()
                }
            )

    return final_output
